﻿import de.alex_uhlmann.animationpackage.APCore;
import de.alex_uhlmann.animationpackage.utility.IVisitor;
import de.alex_uhlmann.animationpackage.utility.IVisitorElement;
import de.alex_uhlmann.animationpackage.utility.Animator;
import de.alex_uhlmann.animationpackage.utility.Pause;
import de.andre_michelle.events.ImpulsDispatcher;
import com.robertpenner.easing.*;
import mx.effects.Tween;

/**
* @class AnimationCore
* @author Alex Uhlmann
* @description  All classes in AnimationPackage that use animations inherit their 
* 			animationStyle properties from AnimationCore. If you don't specify 
* 			the animationStyle properties either with the animationStyle() method 
* 			or directly via the properties, the default animationStyle properties are used.
* 			AnimationCore gives you the opportunity to set and retrieve default properties 
* 			for animationStyle properties. 
* 			Here is how those default properties are defined by default:		
* 			<pre><blockquote>
* 			AnimationCore.duration_def = 1000;
*			AnimationCore.easing_def = Linear.easeNone;
*			AnimationCore.callback_def = null;	
* 			</pre></blockquote>
* 			Note that all default properties have the same name as their instance properties 
* 			but the suffix "_def". For more information about the callback property, take a look at the 
* 			APCore class documentation and the readme.htm > "Event Handling".
* 			<p>
* 			AnimationCore allows you to specify the tween-engine used by AnimationPackage. 
* 			Like discussed in the readme.htm under "Tween-engine" you can choose between time-based tweening 
* 			(INTERVAL), and frame-based tweening (FRAMES). By default time-based tweening is activated. 
* 			You can change the tween mode with AnimationCore.setTweenMode() (see documentation below). 
* 			In INTERVAL mode you control the duration of animations  with milliseconds. In FRAMES mode, by default,
* 			you control the duration of animations with frames. Since this might be a little cumbersome for some developers, 
* 			FRAMES mode can also accept milliseconds if you say so. See AnimationCore.setDurationMode for details.
* 			If you're using frame-based tweening and you're wondering why your animations are so 
* 			slow and sluggish, then check your movie's frame rate.
*  			</p>
* 			<p>
* 			Furthermore, AnimationCore stores functionality that is used by classes that use animations. 
* 			This includes protected functionality to stop, pause, resume, lock and unlock animations 
* 			and static pauseAll and resumeAll methods enable you to pause/resume all running animations.
* 			<p>
* @usage <tt>private class constructor</tt> 
*/
class de.alex_uhlmann.animationpackage.animation.AnimationCore 
													extends APCore 
													implements IVisitorElement{	
	
	/*static stardard properties*/
	/** 
	* @property AnimationCore.duration_def (Number)(static) Duration of animation in milliseconds or frames. Default is milliseconds.
	* @property AnimationCore.easing_def (Function)(static) Easing equation in Robert Penner style. Default equation is Linear.easeNone. www.robertpenner.com/easing/
	* @property AnimationCore.callback_def (String)(static) Function to invoke after animation. See APCore class.	
	*/	
	public static var duration_def:Number = 1000;
	public static var easing_def:Function = Linear.easeNone;
	public static var callback_def:String = null;
	private static var tweenMode:String = "INTERVAL";
	private static var durationMode:String = "MS";
	private static var startTime:Number;
	private static var equivalentsRemoved_static:Boolean = false;
	private var equivalentsRemoved:Boolean = null;
	public var duration:Number;
	public var easing:Function;
	private var easingParams:Array = null;
	public var callback:String;
	private var durationInFrames:Number;
	private var startInitialized:Boolean = false;
	private var startValue:Number;
	private var startValues:Array;
	private var endValue:Number;
	private var endValues:Array;
	private var currentValue:Number;
	private var currentValues:Array;	
	private var mc:MovieClip;
	private var mcs:Array;
	private var rounded:Boolean = false;
	private var forceEndVal:Boolean = true;
	public var myAnimator:Animator;	
	private var locked:Boolean = false;	
	private var myPause:Pause;
	private var tweening:Boolean = false;
	private var paused:Boolean = false;
	private var initAnimation:Function;
	private var invokeAnimation:Function;
	
	private function AnimationCore(noEvents:Boolean) {
		super.init(noEvents);
	}	
	
	/*initialisation for classes with one identifier to animate*/
	private function init():Void {
		if(arguments[0] instanceof Array) {
			var values:Array = arguments[0];
			arguments.shift();
			arguments.unshift(values.pop());				
			this.initAnimation.apply(this, arguments);
			this.setStartValue(values[0]);
		} else if(arguments.length > 0) {
			this.initAnimation.apply(this, arguments);
		}
	}
		
	public function run():Void {		
		this.init.apply(this, arguments);
		this.invokeAnimation(0, 100);		
	}
	
	public function animate(start:Number, end:Number):Void {		
		this.invokeAnimation(start, end);
	}	
	
	public function goto(percentage:Number):Void {
		this.invokeAnimation(percentage);
	}
	
	public function accept(visitor:IVisitor):Void {
		visitor.visit(this);
	}
	
	public function getStartValue(Void):Number {
		return this.startValue;
	}
	
	public function setStartValue(startValue:Number, optional:Boolean):Boolean {
		if(this.startInitialized == false || optional == null) {			
			this.startInitialized = true;			
			this.startValue = startValue;
			return true;
		}
		return false;
	}	
	
	public function getStartValues(Void):Array {
		return this.startValues;
	}
	
	public function setStartValues(startValues:Array, optional:Boolean):Boolean {
		if(this.startInitialized == false || optional != true) {			
			this.startInitialized = true;			
			this.startValues = startValues;			
			return true;
		}
		return false;
	}
	
	public function getEndValue(Void):Number {		
		return this.endValue;
	}
	
	public function setEndValue(endValue:Number):Boolean {
		this.endValue = endValue;
		return true;
	}	
	
	public function getEndValues(Void):Array {		
		return this.endValues;
	}
	
	public function setEndValues(endValues:Array):Boolean {		
		this.endValues = endValues;
		return true;
	}		
	
	public function getCurrentValue(Void):Number {		
		return this.currentValue;
	}
	
	public function getCurrentValues(Void):Array {		
		return this.currentValues;
	}
	
	public function getCurrentPercentage(Void):Number {
		var start:Number;
		var end:Number;
		var current:Number;
		/*
		* if any values is null, AnimationCore assumes 
		* that the animation has not yet started and therefore returns 0%.
		*/
		if(this.startValue == null) {
			/*
			* if start and end values are the same during animation
			* it's impossible to compute the current percentage.
			* Therefore, only compute values that have different start 
			* and end values.
			*/		
			var startValues:Array = this.getStartValues();
			if(startValues == null) {
				return 0;
			}
			var endValues:Array = this.getEndValues();
			if(endValues == null) {
				return 0;
			}
			var i:Number = startValues.length;
			while(--i>-1) {
				if(startValues[i] != endValues[i]) {					
					break;
				}
			}
			start = startValues[i];			
			end = endValues[i];
			if(this.getCurrentValues()[i] != null) {
				current = this.getCurrentValues()[i];
			} else {
				return 0;
			}
		} else {
			if(this.getStartValue() != null) {
				start = this.getStartValue();
			} else {
				return 0;
			}
			if(this.getEndValue() != null) {
				end = this.getEndValue();
			} else {
				return 0;
			}
			if(this.getCurrentValue() != null) {
				current = this.getCurrentValue();
			} else {
				return 0;
			}
		}
		
		if(start < end) {
			return (current - start) / (end - start) * 100;
		} else {
			return 100 - ((current - end) / (start - end) * 100);
		}
	}
	
	public function getDurationElapsed(Void):Number {		
		return this.myAnimator.getDurationElapsed();
	}
	
	public function getDurationRemaining(Void):Number {		
		return this.myAnimator.getDurationRemaining();
	}
	
	/*animationStyle, getter, setter*/
	public function animationStyle(duration:Number, easing:Object, callback:String):Void {                
		var temp;
		if(typeof(easing) == "string") {                        
			temp = easing;                        
			this.callback = temp;
			this.easing = AnimationCore.easing_def;
		} else if (easing == null) {                        
			this.easing = AnimationCore.easing_def;
			this.callback = callback;
		} else if (easing instanceof Array) {
			this.easing = easing[0];
			this.easingParams = easing.slice(1);
			this.callback = callback;
		} else {                        
			temp = easing;
			this.easing = temp;
			this.callback = callback;
		}		
		duration == null ? this.duration = AnimationCore.duration_def : this.duration = duration;
		if(this.callback == null) {
			this.callback = AnimationCore.callback_def
		}		
	}
	
	public function get movieclip():MovieClip {
		return this.mc;
	}
	
	public function set movieclip(mc:MovieClip):Void {
		this.mc = mc;
	}
	
	public function get movieclips():Array {
		return this.mcs;
	}
	
	public function set movieclips(mcs:Array):Void {
		this.mcs = mcs;
	}
	
	public function checkIfRounded(Void):Boolean {
		return this.rounded;
	}
	
	public function roundResult(rounded:Boolean):Void {
		this.rounded = rounded;
	}
	
	public function forceEnd(forceEndVal:Boolean):Void {
		this.forceEndVal = forceEndVal;
	}
	
	public function setOptimizationMode(optimize:Boolean):Void {		
		this.equivalentsRemoved = optimize;
	}
	
	public function getOptimizationMode(Void):Boolean {		
		return this.equivalentsRemoved;
	}

	/**
	* @method setOptimizationMode
	* @description 	Allows to explicitly remove parts of the animation that don't change during 
	* 				the animation. 
	* 				This can add additional performance to your animation. Note that 
	* 				setting this method to true has side effects. If all start and end values match, 
	* 				the animation won't start and will immediatly invoke an onEnd event. 
	* 				The order of values returned by getStartValue(s), getCurrentValue(s), 
	* 				getEndValue(s) and the value property of the eventObject returned 
	* 				by EventDispatcher might change if you set this method to true. You can 
	* 				still retrieve the parts of the animation that are actually animated 
	* 				if you access the Animator instance of your animation class via 
	* 				myAnimator. Ask <code>myInstance.myAnimator.setter</code> to retrieve 
	* 				all currently animated parts of the animation. See Animator 
	* 				documentation. Of cource, if you know your input values you would 
	* 				probably look at them.<p>
	* 				Note that each IAnimatable class offers an instance method named setOptimizationMode 
	* 				(without the "s") that allows you to remove parts only of the 
	* 				animation instance that don't change during the animation.
	* @usage   <pre>myInstance.setOptimizationMode(optimize);</pre>
	* @param optimize (Boolean)
	*/	
	public static function setOptimizationModes(optimize:Boolean):Void {
		AnimationCore.equivalentsRemoved_static = optimize;
	}
	
	public static function getOptimizationModes(Void):Boolean {
		return AnimationCore.equivalentsRemoved_static;
	}
	
	/**
	* @method AnimationCore.getTweenMode
	* @description 	static, returns the current tween mode. 
	* @usage   <tt>AnimationCore.getTweenMode();</tt>
	* @return String that specifies the tween mode. Either "INTERVAL" or "FRAMES".
	*/
	public static function getTweenMode(Void):String {
		return AnimationCore.tweenMode;
	}
	
	/**
	* @method AnimationCore.setTweenMode
	* @description 	static, sets the current tween mode.
	* @usage   <tt>AnimationCore.setTweenMode();</tt> 	
	* @param t (String) Either "INTERVAL" for time-based tweening or "FRAMES" for frame-based tweening.
	* @returns   <code>true</code> if setting tween mode was successful, 
	*                  <code>false</code> if not successful.
	*/
	public static function setTweenMode(t:String):Boolean {
		if(t == "INTERVAL") {
			AnimationCore.tweenMode = t;
			AnimationCore.durationMode = "MS"
		} else if(t == "FRAMES") {
			AnimationCore.tweenMode = t;
			AnimationCore.durationMode = "FRAMES"
		} else {
			return false;
		}
		return true;
	}
	
	/**
	* @method AnimationCore.getDurationMode
	* @description 	static, returns the current duration mode. 
	* @usage   <tt>AnimationCore.getDurationMode();</tt>
	* @return String that specifies the duration mode. Either "MS" or "FRAMES".
	*/
	public static function getDurationMode(Void):String {
		return AnimationCore.durationMode;
	}
	
	/**
	* @method AnimationCore.setDurationMode
	* @description 	static, sets the current duration mode. Only useful in tween mode "FRAMES". Only then 
	* 				you can set the duration mode from "FRAMES", which is the default, to "MS". This allows you 
	* 				to control animations with milliseconds in tween mode FRAMES. Please note that AnimationPackage 
	* 				internally still has to convert your specified milliseconds to frames. This is done by guessing 
	* 				the frame rate (fps = frames per second) of your movie. Therefore, controlling animations with milliseconds in tween mode 
	* 				FRAMES might not be accurate. Since AnimationPackage calculates the fps as soon as you use it 
	* 				for the first time, the example animation below, will start in a one second delay, since this is the time 
	* 				AnimationPackage needs to calculate the fps.
	* 				<p>
	* 				Example 1:
	* 				<blockquote><pre>			
	*				AnimationCore.setTweenMode("FRAMES");
	*				AnimationCore.setDurationMode("MS");
	*				var myRotation:Rotation = new Rotation(mc);
	*				myRotation.animationStyle(1500,Sine.easeInOut);
	*				myRotation.run(360);
	*				</pre></blockquote>
	* 				With <tt>APCore.setFPS()</tt> you can also set the fps manually. In this case or if 
	* 				you have used any class of AnimationPackage earlier (at least one second), then no 
	* 				start up time is required. See APCore for more information 
	* 				about watching, unwatching, setting and getting your movie's frame rate (fps).
	* @usage   <tt>AnimationCore.setDurationMode();</tt> 	
	* @param d (String) Either "MS" for milliseconds or "FRAMES".
	* @returns   <code>true</code> if setting duration mode was successful, 
	*                  <code>false</code> if not successful.
	*/
	public static function setDurationMode(d:String):Boolean {
		if(AnimationCore.tweenMode == "FRAMES" && (d == "MS" || d == "FRAMES")) {
			AnimationCore.durationMode = d;
			return true;
		}
		return false;
	}

	/**
	* @method AnimationCore.pauseAll
	* @description 	static, pauses all running animations whether locked or unlocked.
	* @usage   <tt>AnimationCore.pauseAll();</tt> 	  
	*/
	public static function pauseAll(Void):Void {
		if(AnimationCore.tweenMode == "INTERVAL") {
			clearInterval(Tween.IntervalToken);
			AnimationCore.startTime = getTimer();
		} else if(AnimationCore.tweenMode == "FRAMES") {
			ImpulsDispatcher.pause();
		}
	}
	
	/**
	* @method AnimationCore.resumeAll
	* @description 	static, resumes all running animations whether locked or unlocked. 
	* @usage   <tt>AnimationCore.resumeAll();</tt> 	  
	*/
	public static function resumeAll(Void):Void {		
		if(AnimationCore.tweenMode == "INTERVAL") {
			Tween.IntervalToken = setInterval(Tween.Dispatcher, "DispatchTweens", Tween.Interval);
			var aT:Array = Tween.ActiveTweens;
			var len:Number = aT.length;
			var i:Number;
			for (i=0; i<len; i++) {				
				aT[i].startTime += (getTimer() - AnimationCore.startTime);				
			}			
		} else if(AnimationCore.tweenMode == "FRAMES") {
			ImpulsDispatcher.resume();	
		}	
	}
	
	public function isTweening(Void):Boolean {		
		return this.tweening;
	}
	
	public function isPaused(Void):Boolean {
		return this.paused;
	}
	
	public function stop(Void):Boolean {
		if(this.locked == true || this.tweening == false) {
			this.paused = false;
			return false;
		} else {
			this.tweening = false;
			this.paused = false;
			this.myAnimator.stopMe();
			return true;
		}
	}	
	
	public function pause(duration:Number):Boolean {
		if(this.locked == true || this.tweening == false) {			
			return false;
		} else {			
			this.tweening = false;
			this.paused = true;
			this.myAnimator.paused = true;
			if(duration != null) {				
				if(this.myPause == null) {
					this.myPause = new Pause();
				} else {					
					this.myPause.stop();
				}
				if(AnimationCore.tweenMode == "INTERVAL") {
					this.myPause.waitMS(duration, this, "resume");
				} else if(AnimationCore.tweenMode == "FRAMES") {
					if(AnimationCore.durationMode == "MS") {
						duration = APCore.milliseconds2frames(duration)
					}
					this.myPause.waitFrames(duration, this, "resume");
				}
			}			
			this.myAnimator.pauseMe();
			return true;
		}
	}
	
	public function resume(Void):Boolean {		
		if(this.locked == true || this.myAnimator.paused == false) {
			return false;
		} else {		
			this.tweening = true;
			this.paused = false;
			this.myAnimator.paused = false;
			this.myAnimator.resumeMe();
			return true;
		}
	}	
	
	public function lock(Void):Void {		
		if(this.locked == false) {
			this.locked = true;
		}
	}
	
	public function unlock(Void):Void {		
		if(this.locked == true) {
			this.locked = false;
		}
	}

	public function toString(Void):String {
		return "AnimationCore";
	}
}
